/* corba_client.cpp
 *
 * The entry point for the client program.  This file performs the CORBA setup 
 * and calls client_main() in another file.
 */


// CORBA includes
#include <OB/CORBA.h>
#include <OB/CosNaming.h>
#include <Game.h>

// standard includes
#include <stdlib.h>

// project includes
#include "client.h"

#ifdef HAVE_STD_IOSTREAM
using namespace std;
#endif


int run(CORBA::ORB_ptr orb, int argc, char* argv[]) {
  // Get naming service
  CORBA::Object_var inc;
  try {
    inc = orb->resolve_initial_references("NameService");
  } catch (const CORBA::ORB::InvalidName&) {
    cerr << argv[0] << ": can't resolve `NameService'" <<endl;
    return EXIT_FAILURE;
  }
  if (CORBA::is_nil(inc.in())) {
    cerr << argv[0] << ": `NameService' is a null object reference" <<endl;
    return EXIT_FAILURE;
  }

  CosNaming::NamingContext_var nc = CosNaming::NamingContext::_narrow(inc.in());
  if (CORBA::is_nil(nc.in())) {
    cerr << argv[0] 
         << ": `NameService' is not a NamingContext object reference" << endl;
    return EXIT_FAILURE;
  }

  // Get object by name
  try {
    // Resolve name with the Naming Service
    CosNaming::Name nc1GameServerName;
    nc1GameServerName.length(1);
    nc1GameServerName[0].id = CORBA::string_dup("GameServer");
    CORBA::Object_var gameServerObj = nc->resolve(nc1GameServerName);
    GameServer_var gameServer = GameServer::_narrow(gameServerObj);
    assert(!CORBA::is_nil(gameServer));

    // these values set by join_game
    unsigned int playerId;
    unsigned int gameId;
    Game_var game = gameServer->join_game((CORBA::ULong)playerId,
                                          (CORBA::ULong)gameId);
    assert(!CORBA::is_nil(game));
    printf("playerId = %d gameID = %d\n", playerId, gameId);

    printf("Waiting for other players\n");
    while(1)
      {
	if(game -> is_game_started((CORBA::ULong)playerId,(CORBA::ULong)gameId))
	  break;
	
	//delay as for now
	for(int i=0;i<999999;i++);
      }

    printf("Starting Game\n");

    // start the client code
    int rv = client_main(argc, argv, game, playerId);
    if (rv < 0) {
      cerr << "Unknown error occurred in client_main" << endl;
      return EXIT_FAILURE;
    }
    
    game -> exit_game(playerId,gameId);

  } catch (const CosNaming::NamingContext::NotFound& ex) {
    cerr << argv[0] << ": Got a `NotFound' exception (";
    switch (ex.why) {
      case CosNaming::NamingContext::missing_node:
        cerr << "missing node";
        break;
      case CosNaming::NamingContext::not_context:
        cerr << "not context";
        break;
      case CosNaming::NamingContext::not_object:
        cerr << "not object";
        break;
    }
    cerr << ")" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::CannotProceed&) {
    cerr << argv[0] << ": Got a `CannotProceed' exception" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::InvalidName&) {
    cerr << argv[0] << ": Got an `InvalidName' exception" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::AlreadyBound&) {
    cerr << argv[0] << ": Got an `AlreadyBound' exception" << endl;
    return EXIT_FAILURE;
  } catch (const CosNaming::NamingContext::NotEmpty&) {
    cerr << argv[0] << ": Got a `NotEmpty' exception" << endl;
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}


int main(int argc, char* argv[], char*[]) {
  int status = EXIT_SUCCESS;
  CORBA::ORB_var orb;

  try {
    orb = CORBA::ORB_init(argc, argv, "Game-Client");
    status = run(orb, argc, argv);
  } catch (const CORBA::Exception& ex) {
    cerr << ex << endl;
    status = EXIT_FAILURE;
  }

  if (!CORBA::is_nil(orb)) {
    try {
      orb->destroy();
    } catch (const CORBA::Exception& ex) {
      cerr << ex << endl;
      status = EXIT_FAILURE;
    }
  }

  return status;
}
