/* client.cpp
 *
 * This is an ASCII client for the Borbaman game.
 */


/************
 * INCLUDES *
 ************/ 

// CORBA includes
#include <OB/CORBA.h>
#include <Game.h>

// standard includes
#include <stdlib.h>
#include <ncurses.h>

// C++ includes
#include <vector>
using namespace std;

// project includes
#include "game_constants.h"
#include "client.h"


/*************
 * CONSTANTS *
 *************/

// row offset in the terminal to start displaying the map -- this must be at 
// least 3
const unsigned int MAP_ROW_OFFSET = 5;

// column offset in the terminal to start displaying the map
const unsigned int MAP_COL_OFFSET = 10;

// colors for things on the map
const short CLR_EMPTY    = 1;
const short CLR_BRICK    = 2;
const short CLR_WALL     = 3;
const short CLR_PLAYER   = 4;
const short CLR_BOMB     = 5;
const short CLR_PLAYBOMB = 6;


/***********
 * GLOBALS *
 ***********/

// current game state received from the server
struct GameState gs;

// the player ID for this client
int player_id = 0;

// END FIXME


/**************
 * PROTOTYPES *
 **************/

static int get_player_at(unsigned int x, unsigned int y);
static bool initialize();
static void cleanup();
static void draw_map();
static void draw_map_border();
static void draw_players();
static void draw_bombs();
static void draw_scores();


/***********
 * HELPERS *
 ***********/

// returns the number of the player that is located at (x,y) or -1 if no one
// is at that location
static int get_player_at(unsigned int x, unsigned int y) {
  for (int i = 0; i < NUM_PLAYERS; i++) {
    if ((gs.players[i].x_coord == x) && (gs.players[i].y_coord == y)) {
      return i;
    }
  }
  return -1;
}

// Initializes the client and returns false if successful, otherwise true if an
// error occurred.
static bool initialize() {
  
  // puts the client into curses mode
  initscr();

  // enable colors
  if (start_color() == ERR)
    return true;

  // prevents characters from being echoed to the screen as the user types
  if (noecho() == ERR)
    return true;

  // don't buffer key input
  if (cbreak() == ERR)
    return true;

  // enable extended keys like the arrow keys and function keys
  if (keypad(stdscr, TRUE) == ERR)
    return true;

  // make the cursor invisible
  if (curs_set(0) == ERR)
    return true;

  // create the game colors
  init_pair(CLR_EMPTY,    COLOR_BLACK,  COLOR_BLACK);
  init_pair(CLR_BRICK,    COLOR_YELLOW, COLOR_BLACK);
  init_pair(CLR_WALL,     COLOR_BLACK,  COLOR_WHITE);
  init_pair(CLR_PLAYER,   COLOR_CYAN,   COLOR_BLACK);
  init_pair(CLR_BOMB,     COLOR_WHITE,  COLOR_RED);
  init_pair(CLR_PLAYBOMB, COLOR_CYAN,   COLOR_RED);

  return false;
}

// Performs any necessary cleanup just before the client terminates.
static void cleanup() {
  // restores the terminal to the original mode
  endwin();
}

// Redraws the map on the screen.  This doesn't erase the screen -- it only 
// draws over the map section of the screen.
static void draw_map() {
  // draw the map border
  draw_map_border();

  // draw the map contents
  for (unsigned int row = 0; row < NUM_ROWS; row++) {
    for (unsigned int col = 0; col < NUM_COLS; col++) {
      switch (gs.map_info[row][col]) {
        case MAP_EMPTY:
          attron(COLOR_PAIR(CLR_EMPTY));
          mvaddch(row+MAP_ROW_OFFSET+1, col+MAP_COL_OFFSET+1, ' ');
          attroff(COLOR_PAIR(CLR_EMPTY));
          break;
        case MAP_WALL:
          attron(COLOR_PAIR(CLR_WALL));
          mvaddch(row+MAP_ROW_OFFSET+1, col+MAP_COL_OFFSET+1, '#');
          attroff(COLOR_PAIR(CLR_WALL));
          break;
        case MAP_BRICK:
          attron(COLOR_PAIR(CLR_BRICK));
          mvaddch(row+MAP_ROW_OFFSET+1, col+MAP_COL_OFFSET+1, ACS_CKBOARD);
          attroff(COLOR_PAIR(CLR_BRICK));
          break;
      }
    }
  }
}

// Draws the map border on the screen.
static void draw_map_border() {
  attron(A_BOLD);

  // draw the top and bottom borders
  for (unsigned int col = 1; col < NUM_COLS+1; col++) {
    mvaddch(MAP_ROW_OFFSET, col + MAP_COL_OFFSET, ACS_HLINE);
    mvaddch(MAP_ROW_OFFSET + NUM_ROWS + 1, col + MAP_COL_OFFSET, ACS_HLINE);
  }
  
  // draw the left and right borders
  for (unsigned int row = 1; row < NUM_ROWS+1; row++) {
    mvaddch(row + MAP_ROW_OFFSET, MAP_COL_OFFSET, ACS_VLINE);
    mvaddch(row + MAP_ROW_OFFSET, NUM_COLS + MAP_COL_OFFSET + 1, ACS_VLINE);
  }

  // draw the corners
  mvaddch(MAP_ROW_OFFSET, MAP_COL_OFFSET, ACS_ULCORNER);
  mvaddch(MAP_ROW_OFFSET+NUM_ROWS+1, MAP_COL_OFFSET, ACS_LLCORNER);
  mvaddch(MAP_ROW_OFFSET, MAP_COL_OFFSET+NUM_COLS+1, ACS_URCORNER);
  mvaddch(MAP_ROW_OFFSET+NUM_ROWS+1, MAP_COL_OFFSET+NUM_COLS+1, ACS_LRCORNER);

  attroff(A_BOLD);
}

// Draws the players on the screen.  This should be called after draw_map().
static void draw_players() {
  attron(COLOR_PAIR(CLR_PLAYER) | A_BOLD);
  for (int i = 0; i < NUM_PLAYERS; i++) {
    mvaddch(MAP_ROW_OFFSET + 1 + gs.players[i].y_coord, 
            MAP_COL_OFFSET + 1 + gs.players[i].x_coord, 
            (i + '1'));
  }
  attroff(COLOR_PAIR(CLR_PLAYER) | A_BOLD);
}

// Draws the bombs on the screen.  This should be called after draw_players().
static void draw_bombs() {
  int bcnt;
  int pnum;
  int x, y;
  for (int pcnt = 0; pcnt < NUM_PLAYERS; pcnt++) {
    bcnt = gs.players[pcnt].bombs_dropped;
    for (int i = 0; i < bcnt; i++) {
      // if a player is on the same spot, show the player number with a red
      // background
      x = gs.players[pcnt].bombs[i].x_coord;
      y = gs.players[pcnt].bombs[i].y_coord;
      if ((pnum = get_player_at(x, y)) >= 0) {
        attron(COLOR_PAIR(CLR_PLAYBOMB) | A_BOLD);
        mvaddch(MAP_ROW_OFFSET + 1 + y, MAP_COL_OFFSET + 1 + x, (pnum + '1'));
        attroff(COLOR_PAIR(CLR_PLAYBOMB) | A_BOLD);
      } else {
        attron(COLOR_PAIR(CLR_BOMB));
        mvaddch(MAP_ROW_OFFSET + 1 + y, MAP_COL_OFFSET + 1 + x, 'o');
        attroff(COLOR_PAIR(CLR_BOMB));
      }
    }
  }
}

// Draws the game ID, scores, round number, and time left
static void draw_scores() {
  // game ID
  attron(A_BOLD);
  mvprintw(MAP_ROW_OFFSET - 3, MAP_COL_OFFSET, "Game ID: ");
  attroff(A_BOLD);
  printw("%d", gs.game_id);

  // round number
  attron(A_BOLD);
  mvprintw(MAP_ROW_OFFSET - 2, MAP_COL_OFFSET, "Round: ");
  attroff(A_BOLD);
  printw("%d    ", gs.round_number);

  // time left
  attron(A_BOLD);
  printw("Time: ");
  attroff(A_BOLD);
  printw("%d   ", gs.time_left);

  // scores
  attron(A_BOLD);
  mvprintw(MAP_ROW_OFFSET - 1, MAP_COL_OFFSET, "Scores: ");
  attroff(A_BOLD);
  printw("1=%d  2=%d  3=%d  4=%d",
         gs.players[0].rounds_won, gs.players[1].rounds_won,
         gs.players[2].rounds_won, gs.players[3].rounds_won);
}


/********
 * MAIN *
 ********/

int client_main(int argc, char *argv[], Game_var game, unsigned int playerId) {
  // initialize the client 
  if (initialize()) {
    printf("Error occurred during initialization");
    cleanup();
    return -1;
  }

  mvprintw(20, 0, "press the enter key to exit");

  // refresh the map display on the screen
  draw_map();
  draw_players();
  draw_scores();

  // need to call refresh to update the display with the most recent changes
  refresh();

  // wait for user input
  int c;
  struct ClientAction action;
  action.seq_num   = 0;
  action.player_id = playerId; 
  action.game_id   = game->get_id();

  while ((c = getch()) != '\n') {
    action.seq_num++;
    action.arrow        = MOVE_NONE;
    action.request_bomb = false;

    switch (c) {
      case KEY_LEFT:      // move left
        action.arrow = MOVE_LEFT;
        gs = game->update(action);
        break;
      case KEY_RIGHT:      // move right
        action.arrow = MOVE_RIGHT;
        gs = game->update(action);
        break;
      case KEY_UP:      // move up
        action.arrow = MOVE_UP;
        gs = game->update(action);
        break;
      case KEY_DOWN:      // move down
        action.arrow = MOVE_DOWN;
        gs = game->update(action);
        break;
      case ' ':           // drop bomb
        action.request_bomb = true;
        gs = game->update(action);
        break;
    }

    // redraw the display with the updated game state
    draw_map();
    draw_players();
    draw_bombs();
    draw_scores();

    // update the display
    refresh();
  }

  // perform all cleanup
  cleanup();

  return 0;
}

